from typing import Optional, Iterator

import sqlite3
import zlib
from os import stat
from os.path import isfile
from threading import Lock
from pathlib import Path
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime

from minet.exceptions import SQLArchiveInvalidError
from minet.utils import iterate_over_sqlite_cursor

# Extraction using the sqlite3 command line:
#   $ sqlite3 -Axvf archive.sqlar
#   $ sqlite3 archive.sqlar ".ar -xv --directory archive"

# References:
#  - https://www.sqlite.org/sqlar/doc/trunk/README.md
#  - https://www.sqlite.org/sqlar.html
#  - https://www.sqlite.org/intern-v-extern-blob.html
#  - https://www.sqlite.org/fasterthanfs.html
#  - https://docs.rs/sqlar/latest/sqlar/
#  - https://github.com/Frojdholm/pysqlar

# NOTE: taken from pysqlar:
SQLAR_TABLE_EXPECTED_RESULT = [
    (0, "name", "TEXT", 0, None, 1),
    (1, "mode", "INT", 0, None, 0),
    (2, "mtime", "INT", 0, None, 0),
    (3, "sz", "INT", 0, None, 0),
    (4, "data", "BLOB", 0, None, 0),
]

SQL_PRAGMAS = """
PRAGMA journal_mode=wal;
PRAGMA synchronous=normal;
PRAGMA page_size=16384;
"""

SQL_CREATE = """
CREATE TABLE sqlar (
  name TEXT PRIMARY KEY,  -- name of the file
  mode INT,               -- access permissions
  mtime INT,              -- last modification time
  sz INT,                 -- original file size
  data BLOB               -- compressed content
);
"""

SQL_INSERT = """
INSERT OR REPLACE INTO sqlar (name, mode, mtime, sz, data) VALUES (?, ?, ?, ?, ?);
"""


def sqlar_compress(data: bytes) -> bytes:
    compressed = zlib.compress(data, level=-1)

    if len(compressed) < len(data):
        return compressed

    return data


def sqlar_uncompress(data: bytes, size: int) -> bytes:
    if size == len(data):
        return data

    return zlib.decompress(data)


def is_sqlar(cursor: sqlite3.Cursor) -> bool:
    cursor.execute('PRAGMA table_info("sqlar")')

    return cursor.fetchall() == SQLAR_TABLE_EXPECTED_RESULT


def get_safe_path(name: str) -> str:
    return str(Path(name).as_posix())


@dataclass
class SQLiteArchiveRecord:
    __slots__ = ("name", "mode", "mtime", "size", "data")

    name: str
    mode: int
    mtime: int
    size: int
    data: bytes

    @property
    def is_compressed(self) -> bool:
        return self.size != len(self.data)

    @property
    def uncompressed_data(self) -> bytes:
        return sqlar_uncompress(self.data, self.size)

    @property
    def is_dir(self) -> bool:
        return self.size == 0

    @property
    def is_symlink(self) -> bool:
        return self.size == -1

    @property
    def is_file(self) -> bool:
        return self.size > 0


# NOTE: this class is threadsafe but does not allow for any concurrent access
class SQLiteArchive:
    in_memory: bool
    lock: Lock
    connection: sqlite3.Connection
    default_mode: int

    def __init__(self, filename: Optional[str] = None):
        self.in_memory = False
        self.default_mode = 0o664

        already_exists = False

        if filename is None:
            self.in_memory = True
            filename = ":memory:"
        else:
            already_exists = isfile(filename)

        self.connection = sqlite3.connect(filename, check_same_thread=False)
        self.lock = Lock()

        if not self.in_memory:
            self.default_mode = stat(filename).st_mode & 0o777

        if not already_exists:
            self.connection.executescript(SQL_PRAGMAS)
            self.connection.execute(SQL_CREATE)
            self.connection.commit()
        else:
            with self.transaction() as cursor:
                if not is_sqlar(cursor):
                    raise SQLArchiveInvalidError

    @contextmanager
    def transaction(self):
        cursor = None

        try:
            with self.lock, self.connection:
                cursor = self.connection.cursor()
                yield cursor
        finally:
            if cursor is not None:
                cursor.close()

    def __len__(self) -> int:
        with self.transaction() as cursor:
            cursor.execute("SELECT count(*) FROM sqlar;")
            return cursor.fetchone()[0]

    def write(self, name: str, data: bytes, mtime: Optional[int] = None) -> None:
        with self.transaction() as cursor:
            safe_path = get_safe_path(name)
            compressed_data = sqlar_compress(data)
            mtime = int(datetime.utcnow().timestamp()) if mtime is None else mtime

            cursor.execute(
                SQL_INSERT,
                (safe_path, self.default_mode, mtime, len(data), compressed_data),
            )

    def read(self, name: str) -> SQLiteArchiveRecord:
        with self.transaction() as cursor:
            cursor.execute(
                "SELECT * FROM sqlar WHERE name = ? LIMIT 1;", (get_safe_path(name),)
            )

            row = cursor.fetchone()

            if row is None:
                raise KeyError(name)

            return SQLiteArchiveRecord(*row)

    def __iter__(self) -> Iterator[SQLiteArchiveRecord]:
        with self.transaction() as cursor:
            cursor.execute("SELECT * FROM sqlar;")

            for row in iterate_over_sqlite_cursor(cursor):
                yield SQLiteArchiveRecord(*row)

    def close(self):
        self.connection.close()

    def __del__(self):
        self.close()



from typing import Any

from bs4 import SoupStrainer, BeautifulSoup

from .types import NamedScraper


class TitleScraper(NamedScraper):
    name = "title"
    fieldnames = ["title"]
    plural = False
    output_type = "scalar"
    strainer = SoupStrainer(name="title")

    def scrape(self, soup: BeautifulSoup, context=None) -> Any:
        title_elem = soup.find(name="title")

        if title_elem is None:
            return None

        return title_elem.get_text().strip()



from minet.scrape.classes.base import ScraperBase
from minet.scrape.classes.definition import DefinitionScraper, validate, scrape
from minet.scrape.classes.function import FunctionScraper
from minet.scrape.classes.named import NamedScraper, NAMED_SCRAPERS



# =============================================================================
# Minet CrowdTangle CLI Action
# =============================================================================
#
# Logic of the `ct` action.
#
from casanova import RowCountResumer, LastCellResumer

from minet.cli.argparse import SplitterType, CSVFileType

# TODO: lazyloading issue
from minet.crowdtangle.constants import (
    CROWDTANGLE_DEFAULT_RATE_LIMIT,
    CROWDTANGLE_SORT_TYPES,
    CROWDTANGLE_DEFAULT_START_DATE,
    CROWDTANGLE_SEARCH_FIELDS,
    CROWDTANGLE_SUMMARY_SORT_TYPES,
)

from minet.cli.argparse import command, ConfigAction

FORMAT_ARGUMENT = {
    "flags": ["-f", "--format"],
    "help": "Output format.",
    "choices": ["csv", "jsonl"],
    "default": "csv",
}

CROWDTANGLE_LEADERBOARD_SUBCOMMAND = command(
    "leaderboard",
    "minet.cli.crowdtangle.leaderboard",
    title="Minet CrowdTangle Leaderboard Command",
    description="""
        Gather information and aggregated stats about pages and groups of the designated dashboard (indicated by a given token).

        For more information, see the API endpoint documentation: https://github.com/CrowdTangle/API/wiki/Leaderboard.
    """,
    epilog="""
        Examples:

        . Fetching accounts statistics for every account in your dashboard:
            $ minet ct leaderboard --token YOUR_TOKEN > accounts-stats.csv
    """,
    arguments=[
        FORMAT_ARGUMENT,
        {
            "flags": ["-l", "--limit"],
            "help": "Maximum number of accounts to retrieve. Will fetch every account by default.",
            "type": int,
        },
        {
            "flag": "--list-id",
            "help": "Optional list id from which to retrieve accounts.",
        },
        {
            "flag": "--start-date",
            "help": "The earliest date at which to start aggregating statistics (UTC!). You can pass just a year or a year-month for convenience.",
        },
    ],
)

CROWDTANGLE_LISTS_SUBCOMMAND = command(
    "lists",
    "minet.cli.crowdtangle.lists",
    title="Minet CrowdTangle Lists Command",
    description="""
        Retrieve the lists from a CrowdTangle dashboard (indicated by a given token).

        For more information, see the API endpoint documentation: https://github.com/CrowdTangle/API/wiki/Lists.
    """,
    epilog="""
        Examples:

        . Fetching a dashboard's lists:
            $ minet ct lists --token YOUR_TOKEN > lists.csv
    """,
)

CROWDTANGLE_POSTS_BY_ID_SUBCOMMAND = command(
    "posts-by-id",
    "minet.cli.crowdtangle.posts_by_id",
    title="Minet CrowdTangle Post By Id Command",
    description="""
        Retrieve metadata about batches of posts using Crowdtangle's API.

        For more information, see the API endpoint documentation: https://github.com/CrowdTangle/API/wiki/Posts#get-postid.
    """,
    epilog="""
        Examples:

        . Retrieving information about a batch of posts:
            $ minet ct posts-by-id post-url -i posts.csv --token YOUR_TOKEN > metadata.csv

        . Retrieving information about a single post:
            $ minet ct posts-by-id 1784333048289665 --token YOUR_TOKEN
    """,
    variadic_input={
        "dummy_column": "post_url_or_id",
        "item_label": "URL or id",
        "item_label_plural": "URLs or ids",
    },
    resumer=RowCountResumer,
)

CROWDTANGLE_POSTS_SUBCOMMAND = command(
    "posts",
    "minet.cli.crowdtangle.posts",
    title="Minet CrowdTangle Posts Command",
    description="""
        Gather post data from the designated dashboard (indicated by a given token).

        For more information, see the API endpoint documentation: https://github.com/CrowdTangle/API/wiki/Posts.
    """,
    epilog="""
        Examples:

        . Fetching the 500 most latest posts from a dashboard (a start date must be precised):
            $ minet ct posts --token YOUR_TOKEN --limit 500 --start-date 2021-01-01 > latest-posts.csv

        . If your collection is interrupted, it can be restarted from the last data collected with the --resume option:
            $ minet ct posts --token YOUR_TOKEN --limit 500 --start-date 2021-01-01 --resume --output latest-posts.csv

        . Fetching all the posts from a specific list of groups or pages:
            $ minet ct posts --token YOUR_TOKEN --start-date 2021-01-01 --list-ids YOUR_LIST_ID > posts_from_one_list.csv

        To know the different list ids associated with your dashboard:
            $ minet ct lists --token YOUR_TOKEN
    """,
    resumer=LastCellResumer,
    resumer_kwargs={"value_column": "datetime"},
    arguments=[
        {
            "flag": "--chunk-size",
            "help": "When sorting by date (default), the number of items to retrieve before shifting the initial query to circumvent the APIs limitations.",
            "type": int,
            "default": 500,
        },
        {
            "flag": "--end-date",
            "help": "The latest date at which a post could be posted (UTC!). You can pass just a year or a year-month for convenience.",
        },
        FORMAT_ARGUMENT,
        {
            "flag": "--language",
            "help": "Language of posts to retrieve.",
        },
        {
            "flags": ["-l", "--limit"],
            "help": "Maximum number of posts to retrieve. Will fetch every post by default.",
            "type": int,
        },
        {
            "flag": "--list-ids",
            "help": "Ids of the lists from which to retrieve posts, separated by commas.",
            "type": SplitterType(),
        },
        {
            "flag": "--sort-by",
            "help": "The order in which to retrieve posts.",
            "choices": CROWDTANGLE_SORT_TYPES,
            "default": "date",
        },
        {
            "flag": "--start-date",
            "help": "The earliest date at which a post could be posted (UTC!). You can pass just a year or a year-month for convenience.",
            "default": CROWDTANGLE_DEFAULT_START_DATE,
        },
    ],
)

CROWDTANGLE_SEARCH_SUBCOMMAND = command(
    "search",
    "minet.cli.crowdtangle.search",
    title="Minet CrowdTangle Search Command",
    description="""
        Search posts on the whole CrowdTangle platform.

        For more information, see the API endpoint documentation: https://github.com/CrowdTangle/API/wiki/Search.
    """,
    epilog="""
        Examples:

        . Fetching all the 2021 posts containing the words 'acetylsalicylic acid':
            $ minet ct search 'acetylsalicylic acid' --start-date 2021-01-01 --token YOUR_TOKEN > posts.csv
    """,
    arguments=[
        {"name": "terms", "help": "The search query term or terms."},
        {
            "flag": "--and",
            "help": "AND clause to add to the query terms.",
        },
        {
            "flag": "--chunk-size",
            "help": "When sorting by date (default), the number of items to retrieve before shifting the initial query to circumvent the APIs limitations.",
            "type": int,
            "default": 500,
        },
        {
            "flag": "--end-date",
            "help": "The latest date at which a post could be posted (UTC!). You can pass just a year or a year-month for convenience.",
        },
        FORMAT_ARGUMENT,
        {
            "flag": "--in-list-ids",
            "help": "Ids of the lists in which to search, separated by commas.",
            "type": SplitterType(),
        },
        {
            "flag": "--language",
            "help": 'Language ISO code like "fr" or "zh-CN".',
        },
        {
            "flags": ["-l", "--limit"],
            "help": "Maximum number of posts to retrieve. Will fetch every post by default.",
            "type": int,
        },
        {
            "flag": "--not-in-title",
            "help": "Whether to search terms in account titles also.",
            "action": "store_true",
        },
        {"flag": "--offset", "help": "Count offset.", "type": int},
        {
            "flags": ["-p", "--platforms"],
            "help": "The platforms from which to retrieve links (facebook, instagram, or reddit). This value can be comma-separated.",
            "type": SplitterType(),
        },
        {
            "flag": "--search-field",
            "help": "In what to search the query.",
            "choices": CROWDTANGLE_SEARCH_FIELDS,
            "default": "text_fields_and_image_text",
        },
        {
            "flag": "--sort-by",
            "help": "The order in which to retrieve posts.",
            "choices": CROWDTANGLE_SORT_TYPES,
            "default": "date",
        },
        {
            "flag": "--start-date",
            "help": "The earliest date at which a post could be posted (UTC!). You can pass just a year or a year-month for convenience.",
            "default": CROWDTANGLE_DEFAULT_START_DATE,
        },
        {
            "flag": "--types",
            "help": "Types of post to include, separated by comma.",
            "type": SplitterType(),
        },
    ],
)

CROWDTANGLE_SUMMARY_SUBCOMMAND = command(
    "summary",
    "minet.cli.crowdtangle.summary",
    title="Minet CrowdTangle Link Summary Command",
    description="""
        Retrieve aggregated statistics about link sharing on the Crowdtangle API and by platform.

        For more information, see the API endpoint documentation: https://github.com/CrowdTangle/API/wiki/Links.
    """,
    epilog="""
        Examples:

        . Computing a summary of aggregated stats for urls contained in a CSV row:
            $ minet ct summary url -i urls.csv --token YOUR_TOKEN --start-date 2019-01-01 > summary.csv
    """,
    variadic_input={"dummy_column": "url", "item_label": "URL"},
    arguments=[
        {
            "flags": ["-p", "--platforms"],
            "help": "The platforms from which to retrieve links (facebook, instagram, or reddit). This value can be comma-separated.",
            "type": SplitterType(),
        },
        {
            "name": "--posts",
            "help": "Path to a file containing the retrieved posts.",
            "type": CSVFileType("w"),
        },
        {
            "flag": "--sort-by",
            "help": "How to sort retrieved posts.",
            "choices": CROWDTANGLE_SUMMARY_SORT_TYPES,
            "default": "date",
        },
        {
            "flag": "--start-date",
            "help": "The earliest date at which a post could be posted (UTC!). You can pass just a year or a year-month for convenience.",
            "default": CROWDTANGLE_DEFAULT_START_DATE,
            "required": True,
        },
    ],
)


CROWDTANGLE_COMMAND = command(
    "crowdtangle",
    "minet.cli.crowdtangle",
    "Minet Crowdtangle Command",
    aliases=["ct"],
    description="""
        Gather data from the CrowdTangle APIs easily and efficiently.
    """,
    common_arguments=[
        {
            "flag": "--rate-limit",
            "help": "Authorized number of hits by minutes.",
            "type": int,
            "default": CROWDTANGLE_DEFAULT_RATE_LIMIT,
            "rc_key": ["crowdtangle", "rate_limit"],
            "action": ConfigAction,
        },
        {
            "flags": ["-t", "--token"],
            "help": "CrowdTangle dashboard API token. Rcfile key: crowdtangle.token",
            "action": ConfigAction,
            "rc_key": ["crowdtangle", "token"],
            "required": True,
        },
    ],
    subcommands=[
        CROWDTANGLE_LEADERBOARD_SUBCOMMAND,
        CROWDTANGLE_LISTS_SUBCOMMAND,
        CROWDTANGLE_POSTS_BY_ID_SUBCOMMAND,
        CROWDTANGLE_POSTS_SUBCOMMAND,
        CROWDTANGLE_SEARCH_SUBCOMMAND,
        CROWDTANGLE_SUMMARY_SUBCOMMAND,
    ],
)



# =============================================================================
# Minet BuzzSumo Limit CLI Action
# =============================================================================
#
# Logic of the `bz limit` action.
#
from ebbe import format_int

from minet.buzzsumo import BuzzSumoAPIClient
from minet.cli.console import console
from minet.cli.buzzsumo.utils import with_buzzsumo_fatal_errors


@with_buzzsumo_fatal_errors
def action(cli_args):
    client = BuzzSumoAPIClient(cli_args.token)

    limit = client.limit()

    console.log_with_time(
        "With your token, you can still make",
        "[info]%s[/info]" % format_int(limit),
        "calls to the BuzzSumo API until the end of the month.",
    )



from typing import List, Set

from dataclasses import dataclass
from casanova import TabularRecord
from ebbe import getpath

from minet.instagram.utils import (
    extract_hashtags,
    extract_handles,
    short_code_to_url,
)
from minet.instagram.constants import INSTAGRAM_MEDIA_TYPE
from minet.dates import timestamp_to_isoformat

INSTAGRAM_PATH_TO_MEDIA_IMAGE = ["image_versions2", "candidates", 0, "url"]
INSTAGRAM_PATH_TO_MEDIA_VIDEO = ["video_versions", 0, "url"]

# NOTE: we are lacking a great deal of optionals there...


@dataclass
class InstagramComment(TabularRecord):
    id: str
    index: str
    parent_id: str
    child_index: str
    text: str
    hashtags: List[str]
    mentioned_names: List[str]
    has_translation: bool
    comment_like_count: int
    is_liked_by_media_owner: bool
    child_comment_count: int
    created_at_utc: str
    utc_time: str
    username: str
    user_id: str
    user_full_name: str
    user_is_private: bool
    user_is_mentionable: bool
    user_is_verified: bool
    user_profile_pic_id: str
    user_profile_pic_url: str
    user_fbid_v2: str

    @classmethod
    def from_payload(cls, payload) -> "InstagramComment":
        text = payload.get("text")

        hashtags = []
        mentioned_names = []

        if payload.get("text"):
            hashtags = extract_hashtags(text)
            mentioned_names = extract_handles(text)

        return cls(
            payload.get("pk"),
            payload.get("comment_index"),
            payload.get("parent_comment_id"),
            payload.get("child_comment_index"),
            payload.get("text"),
            hashtags,
            mentioned_names,
            payload.get("has_translation", False),
            payload.get("comment_like_count"),
            payload.get("is_liked_by_media_owner"),
            payload.get("child_comment_count"),
            payload.get("created_at_utc"),
            timestamp_to_isoformat(payload.get("created_at_utc")),
            getpath(payload, ["user", "username"]),
            payload.get("user_id"),
            getpath(payload, ["user", "full_name"]),
            getpath(payload, ["user", "is_private"]),
            getpath(payload, ["user", "is_mentionable"]),
            getpath(payload, ["user", "is_verified"]),
            getpath(payload, ["user", "profile_pic_id"]),
            getpath(payload, ["user", "profile_pic_url"]),
            getpath(payload, ["user", "fbid_v2"]),
        )


@dataclass
class InstagramHashtagPost(TabularRecord):
    id: str
    media_type: str
    shortcode: str
    url: str
    caption: str
    hashtags: List[str]
    mentioned_names: List[str]
    liked_by_count: int
    comments_disabled: bool
    comment_count: int
    preview_like_count: int
    video_view_count: int
    owner_id: str
    taken_at_timestamp: str
    utc_time: str
    accessibility_caption: str
    display_url: str

    @classmethod
    def from_payload(cls, payload) -> "InstagramHashtagPost":
        text = getpath(payload, ["edge_media_to_caption", "edges", 0, "node", "text"])

        hashtags = []
        mentioned_names = []

        if text:
            hashtags = extract_hashtags(text)
            mentioned_names = extract_handles(text)

        return cls(
            payload["id"],
            payload["__typename"],
            payload["shortcode"],
            short_code_to_url(payload["shortcode"]),
            text,
            hashtags,
            mentioned_names,
            payload["edge_liked_by"]["count"],
            payload["comments_disabled"],
            payload["edge_media_to_comment"]["count"],
            payload["edge_media_preview_like"]["count"],
            payload.get("video_view_count"),
            payload["owner"]["id"],
            payload["taken_at_timestamp"],
            timestamp_to_isoformat(payload["taken_at_timestamp"]),
            payload["accessibility_caption"],
            short_code_to_url(payload["shortcode"]) + "media/?size=l",
        )


def get_usertags(item):
    users = getpath(item, ["usertags", "in"])
    usertags = set()
    if not users:
        return usertags
    for user in users:
        username = getpath(user, ["user", "username"])
        usertags.add(username)
    return usertags


@dataclass
class InstagramPost(TabularRecord):
    username: str
    full_name: str
    is_verified: bool
    id: str
    media_type: str
    shortcode: str
    url: str
    caption: str
    hashtags: List[str]
    mentioned_names: List[str]
    like_and_view_counts_disabled: bool
    like_count: int
    comment_count: int
    main_thumbnail_url: str
    medias_type: List[str]
    medias_url: List[str]
    usertags_medias: Set[str]
    taken_at_timestamp: str
    utc_time: str
    coauthor_usernames: List[str]

    @classmethod
    def from_payload(cls, payload) -> "InstagramPost":
        media_type = INSTAGRAM_MEDIA_TYPE.get(payload["media_type"])
        if media_type is None:
            media_type = payload["media_type"]

        text = getpath(payload, ["caption", "text"])

        hashtags = []
        mentioned_names = []

        if text:
            hashtags = extract_hashtags(text)
            mentioned_names = extract_handles(text)

        medias_url = []
        medias_type = []
        usertags_medias = set()

        if media_type == "GraphSidecar":
            carousel = getpath(payload, ["carousel_media"])
            for media in carousel:
                medias_type.append(
                    INSTAGRAM_MEDIA_TYPE.get(getpath(media, ["media_type"]))
                )

                if medias_type[-1] == "GraphImage":
                    medias_url.append(getpath(media, INSTAGRAM_PATH_TO_MEDIA_IMAGE))
                    item_usertags = get_usertags(media)
                    usertags_medias = usertags_medias.union(item_usertags)

                elif medias_type[-1] == "GraphVideo":
                    medias_url.append(getpath(media, INSTAGRAM_PATH_TO_MEDIA_VIDEO))
                    item_usertags = get_usertags(media)
                    usertags_medias = usertags_medias.union(item_usertags)

        elif media_type == "GraphImage":
            medias_type.append("GraphImage")
            medias_url.append(getpath(payload, INSTAGRAM_PATH_TO_MEDIA_IMAGE))
            item_usertags = get_usertags(payload)
            usertags_medias = usertags_medias.union(item_usertags)

        elif media_type == "GraphVideo":
            medias_type.append("GraphVideo")
            medias_url.append(getpath(payload, INSTAGRAM_PATH_TO_MEDIA_VIDEO))
            item_usertags = get_usertags(payload)
            usertags_medias = usertags_medias.union(item_usertags)

        coauthor_usernames = set()

        for other_user in payload.get("coauthor_producers", []) + payload.get(
            "invited_coauthor_producers", []
        ):
            coauthor_username = other_user.get("username")

            if coauthor_username is not None:
                coauthor_usernames.add(coauthor_username)

        return cls(
            getpath(payload, ["user", "username"]),
            getpath(payload, ["user", "full_name"]),
            getpath(payload, ["user", "is_verified"]),
            payload["pk"],
            media_type,
            payload["code"],
            short_code_to_url(payload["code"]),
            text,
            hashtags,
            mentioned_names,
            payload["like_and_view_counts_disabled"],
            payload["like_count"],
            payload["comment_count"],
            short_code_to_url(payload["code"]) + "media/?size=l",
            medias_type,
            medias_url,
            usertags_medias,
            payload["taken_at"],
            timestamp_to_isoformat(payload["taken_at"]),
            list(coauthor_usernames),
        )


@dataclass
class InstagramUser(TabularRecord):
    id: str
    username: str
    full_name: str
    is_private: bool
    profile_pic_url: str
    profile_pic_id: str
    is_verified: bool
    has_anonymous_profile_picture: str
    has_highlight_reels: bool
    account_badges: List[str]
    similar_user_id: str
    latest_reel_media: str
    is_favorite: bool

    @classmethod
    def from_payload(cls, payload) -> "InstagramUser":
        return cls(
            payload.get("pk"),
            payload.get("username"),
            payload.get("full_name"),
            payload.get("is_private"),
            payload.get("profile_pic_url"),
            payload.get("profile_pic_id"),
            payload.get("is_verified"),
            payload.get("has_anonymous_profile_picture"),
            payload.get("has_highlight_reels"),
            payload.get("account_badges"),
            payload.get("similar_user_id"),
            payload.get("latest_reel_media"),
            payload.get("is_favorite"),
        )


@dataclass
class InstagramUserInfo(TabularRecord):
    username: str
    id: str
    full_name: str
    profile_pic_url_hd: str
    is_private: bool
    is_verified: bool
    biography: str
    bio_links_title: List[str]
    bio_links_url: List[str]
    biography_with_username: List[str]
    biography_with_hashtag: List[str]
    follower_count: str
    follow_count: str
    hide_like_and_view_counts: bool
    pronouns: List[str]
    has_ar_effects: bool
    has_clips: bool
    has_guides: bool
    has_channel: bool
    is_business_account: bool
    is_professional_account: bool
    is_supervision_enabled: bool
    is_guardian_of_viewer: bool
    is_supervised_by_viewer: bool
    is_supervised_user: bool
    guardian_id: str
    is_joined_recently: str
    business_category_name: str
    category_name: str
    connected_fb_page: str

    @classmethod
    def from_payload(cls, payload) -> "InstagramUserInfo":
        bio_links_title = []
        bio_links_url = []
        biography_with_username = []
        biography_with_hashtag = []

        bio_links = payload.get("bio_links")
        if bio_links:
            for link in bio_links:
                bio_links_title.append(link.get("title"))
                bio_links_url.append(link.get("url"))

        biography_with_entities = getpath(
            payload, ["biography_with_entities", "entities"]
        )

        if biography_with_entities:
            for entity in biography_with_entities:
                username = getpath(entity, ["user", "username"])
                if username:
                    biography_with_username.append(username)

                hashtag = getpath(entity, ["user", "hashtag"])
                if hashtag:
                    biography_with_hashtag.append(hashtag)

        pronouns = payload.get("pronouns")

        return cls(
            payload.get("username"),
            payload.get("id"),
            payload.get("full_name"),
            payload.get("profile_pic_url_hd"),
            payload.get("is_private"),
            payload.get("is_verified"),
            payload.get("biography"),
            bio_links_title,
            bio_links_url,
            biography_with_username,
            biography_with_hashtag,
            getpath(payload, ["edge_followed_by", "count"]),
            getpath(payload, ["edge_follow", "count"]),
            payload.get("hide_like_and_view_counts"),
            pronouns,
            payload.get("has_ar_effects"),
            payload.get("has_clips"),
            payload.get("has_guides"),
            payload.get("has_channel"),
            payload.get("is_business_account"),
            payload.get("is_professional_account"),
            payload.get("is_supervision_enabled"),
            payload.get("is_guardian_of_viewer"),
            payload.get("is_supervised_by_viewer"),
            payload.get("is_supervised_user"),
            payload.get("guardian_id"),
            payload.get("is_joined_recently"),
            payload.get("business_category_name"),
            payload.get("category_name"),
            payload.get("connected_fb_page"),
        )



# =============================================================================
# Minet Url Join CLI Action
# =============================================================================
#
# Logic of the `url-join` action.
#
import casanova
from ural.lru import NormalizedLRUTrie
from ebbe import format_int

from minet.cli.console import console
from minet.cli.loading_bar import LoadingBar


def action(cli_args):
    left_reader = casanova.reader(cli_args.input1)
    left_headers = left_reader.fieldnames
    left_idx = None

    assert left_headers is not None

    if cli_args.select:
        assert left_reader.headers is not None

        left_idx = left_reader.headers.select(cli_args.select)
        left_headers = [left_headers[i] for i in left_idx]

    # Applying column prefix now
    left_headers = [cli_args.match_column_prefix + h for h in left_headers]

    right_enricher = casanova.enricher(
        cli_args.input2, cli_args.output, add=left_headers
    )

    # First step is to index left file
    trie = NormalizedLRUTrie()

    with LoadingBar(
        title="Indexing first file", unit="lines", total=left_reader.total
    ) as loading_bar:
        for row, cell in left_reader.cells(cli_args.column1, with_rows=True):
            with loading_bar.step():
                if left_idx is not None:
                    row = [row[i] for i in left_idx]

                urls = [cell]

                if cli_args.separator is not None:
                    urls = cell.split(cli_args.separator)

                for url in urls:
                    url = url.strip()

                    if not url:
                        continue

                    try:
                        trie.set(url, row)
                    except Exception:
                        loading_bar.inc_stat("invalid-url", style="error")
                        continue

    console.print("Indexed [cyan]{}[/cyan] prefixes.\n".format(format_int(len(trie))))

    with LoadingBar(
        title="Matching lines in second file", unit="lines", total=right_enricher.total
    ) as loading_bar:
        for row, url in right_enricher.cells(cli_args.column2, with_rows=True):
            with loading_bar.step():
                url = url.strip()

                match = None

                if url:
                    try:
                        match = trie.match(url)
                    except Exception:
                        loading_bar.inc_stat("invalid-url", style="error")

                if match is None:
                    right_enricher.writerow(row)
                    continue

                right_enricher.writerow(row, match)



# =============================================================================
# Minet Scrape CLI Action
# =============================================================================
#
# Logic of the scrape action.
#
from typing import Optional, Any, List, Dict, Iterator

import casanova
import casanova.ndjson as ndjson
from dataclasses import dataclass
from itertools import count
from threading import Lock
from os.path import basename, isdir

from minet.utils import import_target
from minet.scrape.classes import (
    NAMED_SCRAPERS,
    ScraperBase,
    DefinitionScraper,
    FunctionScraper,
)
from minet.multiprocessing import LazyPool
from minet.exceptions import (
    DefinitionInvalidFormatError,
    GenericModuleNotFoundError,
    TargetInGenericModuleNotFoundError,
)
from minet.scrape.exceptions import (
    InvalidScraperError,
    CSSSelectorTooComplex,
    ScraperEvalError,
    ScraperEvalTypeError,
    ScraperEvalNoneError,
)
from minet.cli.reporters import (
    report_scraper_validation_errors,
    report_scraper_evaluation_error,
)
from minet.serialization import serialize_error_as_slug
from minet.fs import read_potentially_gzipped_path
from minet.cli.console import console
from minet.cli.loading_bar import LoadingBar
from minet.cli.utils import create_fetch_like_report_iterator, FetchReportLikeItem
from minet.cli.exceptions import FatalError


@dataclass
class ScrapeWorkerPayload:
    id: int
    row: List[str]
    path: Optional[str]
    encoding: Optional[str]
    text: Optional[str]
    url: Optional[str]


@dataclass
class ScrapeResult:
    id: int
    error: Optional[Exception] = None
    items: Optional[Any] = None


SCRAPER: Optional[ScraperBase] = None
HEADERS: Optional[casanova.headers] = None


def init_process(scraper: ScraperBase, fieldnames: List[str]):
    global SCRAPER
    global HEADERS

    SCRAPER = scraper
    HEADERS = casanova.headers(fieldnames)


def worker(payload: ScrapeWorkerPayload) -> ScrapeResult:
    assert SCRAPER is not None
    assert HEADERS is not None

    text = payload.text

    result = ScrapeResult(id=payload.id)

    # Reading file
    if text is None:
        try:
            text = read_potentially_gzipped_path(
                payload.path, encoding=payload.encoding, fallback_encoding="utf-8"
            )
        except (FileNotFoundError, UnicodeDecodeError) as e:
            result.error = e
            return result

    # Building context
    row = HEADERS.wrap(payload.row, transient=True)
    context: Dict[str, Any] = {"row": row}

    if payload.url:
        context["url"] = payload.url

    if payload.path:
        context["path"] = payload.path
        context["basename"] = basename(payload.path)

    # Attempting to scrape
    items = SCRAPER.items(text, context=context)

    # NOTE: errors might be raised when we consume the generators created above
    try:
        items = list(items)
    except (ScraperEvalError, ScraperEvalTypeError, ScraperEvalNoneError) as error:
        result.error = error
        return result

    result.items = items

    return result


def action(cli_args):
    # Parsing scraper definition
    try:
        if cli_args.module:
            fn = import_target(cli_args.scraper, default="scrape")
            scraper = FunctionScraper(fn, strain=cli_args.strain)
        elif cli_args.eval:
            scraper = FunctionScraper(cli_args.scraper, strain=cli_args.strain)
        elif cli_args.scraper in NAMED_SCRAPERS:
            scraper = NAMED_SCRAPERS[cli_args.scraper]()
        else:
            scraper = DefinitionScraper(cli_args.scraper, strain=cli_args.strain)

    except GenericModuleNotFoundError:
        raise FatalError(
            [
                "Could not import %s!" % cli_args.scraper,
                "Are you sure the module exists?",
            ]
        )

    except TargetInGenericModuleNotFoundError as e:
        raise FatalError(
            [
                "Could not find the %s target in the %s module!" % (e.name, e.path),
                "Are you sure this class/function/variable exists in the module?",
            ]
        )

    except DefinitionInvalidFormatError:
        raise FatalError(
            [
                "Unknown scraper format!",
                "It should be a JSON or YAML file.",
                "Or did you forget the -m/--module flag?",
            ]
        )

    except FileNotFoundError:
        raise FatalError("Could not find scraper file!")

    except InvalidScraperError as error:
        raise FatalError(
            [
                "Your scraper is invalid! You need to fix the following errors:\n",
                report_scraper_validation_errors(error.validation_errors),
            ]
        )

    except CSSSelectorTooComplex:
        raise FatalError(
            [
                "Your strainer's CSS selector [info]%s[/info] is too complex."
                % cli_args.strain,
                "You cannot use relations to create a strainer.",
                "Try to simplify the selector you passed to --strain.",
            ]
        )

    if not scraper.tabular and cli_args.format == "csv":
        raise FatalError(
            [
                "Your scraper does not yield tabular data.",
                'Try changing it or setting --format to "jsonl".',
            ]
        )

    if cli_args.input_dir is not None and not isdir(cli_args.input_dir):
        raise FatalError(
            'Could not find the [cyan]-I/--input-dir "{}"[/cyan] directory!'.format(
                cli_args.input_dir
            )
        )

    writer_lock = Lock()

    if cli_args.format == "csv":
        if isinstance(scraper, FunctionScraper):
            enricher = casanova.inferring_enricher(
                cli_args.input,
                cli_args.output,
                total=cli_args.total,
                plural_separator=cli_args.plural_separator,
                select=cli_args.select,
                mapping_sample_size=512,
                buffer_optionals=True,
            )
            reader = enricher

            def writerow(row, item):
                enricher.writerow(row, item)

        else:
            assert scraper.fieldnames is not None

            serializer = casanova.CSVSerializer(
                plural_separator=cli_args.plural_separator
            )

            output_fieldnames = scraper.fieldnames

            if cli_args.scraped_column_prefix is not None:
                output_fieldnames = [
                    cli_args.scraped_column_prefix + h for h in output_fieldnames
                ]

            enricher = casanova.enricher(
                cli_args.input,
                cli_args.output,
                total=cli_args.total,
                select=cli_args.select,
                add=output_fieldnames,
            )
            reader = enricher

            def writerow(row, item):
                assert scraper.fieldnames is not None

                if item is None:
                    enricher.writerow(row)
                    return

                if isinstance(item, dict):
                    item = [item.get(f) for f in scraper.fieldnames]
                else:
                    item = [item]

                enricher.writerow(row, (serializer(v) for v in item))  # type: ignore

    else:
        # TODO: casanova should probably expose some ndjson enricher
        reader = casanova.reader(cli_args.input, total=cli_args.total)
        writer = ndjson.writer(cli_args.output)

        def writerow(row, item):
            writer.writerow(item)

    with LoadingBar(
        "Scraping", unit="pages", total=reader.total if not cli_args.glob else None
    ) as loading_bar:
        items = create_fetch_like_report_iterator(cli_args, reader)

        worked_on: Dict[int, FetchReportLikeItem] = {}
        worked_on_lock = Lock()

        # Fun fact: Pool.imap consumes the given iterator in a separate thread...
        def payloads() -> Iterator[ScrapeWorkerPayload]:
            current_id = count()

            for item in items:
                # Items we cannot process
                if item.error is not None:
                    loading_bar.advance()
                    loading_bar.inc_stat(item.error, style="error")

                    # NOTE: we emit an empty line on error if scraper is singular
                    if is_singular:
                        with writer_lock:
                            writerow(item.row, None)

                    continue

                item_id = next(current_id)

                with worked_on_lock:
                    worked_on[item_id] = item

                yield ScrapeWorkerPayload(
                    id=item_id,
                    encoding=item.encoding,
                    path=item.path,
                    text=item.text,
                    row=item.row,
                    url=item.url,
                )

        pool = LazyPool(
            cli_args.processes,
            initializer=init_process,
            initargs=(scraper, reader.fieldnames),
        )

        loading_bar.append_to_title(" (p=%i)" % pool.processes)

        warned_about_input_dir = False

        is_singular = scraper.singular

        with pool:
            for result in pool.imap(
                worker,
                payloads(),
                chunksize=cli_args.chunk_size,
                unordered=cli_args.unordered,
            ):
                with loading_bar.step():
                    with worked_on_lock:
                        original_item = worked_on.pop(result.id)

                    if result.error is not None:
                        # NOTE: we emit an empty line on error if scraper is singular
                        if is_singular:
                            with writer_lock:
                                writerow(original_item.row, None)

                        if isinstance(
                            result.error,
                            (
                                ScraperEvalError,
                                ScraperEvalTypeError,
                                ScraperEvalNoneError,
                            ),
                        ):
                            loading_bar.print(
                                report_scraper_evaluation_error(result.error), end=""
                            )
                        else:
                            error_code = serialize_error_as_slug(result.error)
                            loading_bar.inc_stat(error_code, style="error")

                            if (
                                not warned_about_input_dir
                                and error_code == "file-not-found"
                            ):
                                warned_about_input_dir = True

                                if cli_args.input_dir is None:
                                    console.warning(
                                        "did you forget to give [cyan]-I/--input-dir[/cyan]? "
                                    )
                                else:
                                    console.warning(
                                        'are you sure [cyan]-I/--input-dir "{}"[/cyan] is the correct path?'.format(
                                            cli_args.input_dir
                                        )
                                    )

                        continue

                    # No errors
                    assert result.items is not None
                    items = result.items

                    with writer_lock:
                        for item in items:
                            writerow(original_item.row, item)

                        # NOTE: we emit an empty line for singular scraper if no match occurred
                        if is_singular and not items:
                            writerow(original_item.row, None)

                    loading_bar.inc_stat(
                        "scraped-items", count=len(items), style="info"
                    )



import re
from typing import List, Optional, cast, overload

import warnings
from contextlib import contextmanager
from bs4 import Tag, BeautifulSoup, SoupStrainer

from minet.scrape.std import get_display_text

try:
    from bs4 import XMLParsedAsHTMLWarning
except ImportError:
    XMLParsedAsHTMLWarning = None

WHITESPACE_RE = re.compile(r"\s+")


class SelectionError(Exception):
    pass


def extract(elem: "MinetTag", target: Optional[str]) -> Optional[str]:
    if target is None or target == "text":
        return elem.get_text()

    if target == "html" or target == "inner_html":
        return elem.get_html()

    if target == "outer_html":
        return elem.get_outer_html()

    if target == "display_text":
        return elem.get_display_text()

    return cast(Optional[str], elem.get(target))


class MinetTag(Tag):
    def force_select_one(self, css: str, *args, **kwargs) -> "MinetTag":
        elem = super().select_one(css, *args, **kwargs)

        if elem is None:
            raise SelectionError(css)

        return cast(MinetTag, elem)

    def select_one(self, css: str, *args, **kwargs) -> Optional["MinetTag"]:
        return cast(Optional["MinetTag"], super().select_one(css, *args, **kwargs))

    def select(self, css: str, *args, **kwargs) -> List["MinetTag"]:
        css = css.replace(":contains(", ":-soup-contains(")

        return cast(List["MinetTag"], super().select(css, *args, **kwargs))

    def scrape_one(
        self, css: str, target: Optional[str] = None, strict: bool = False
    ) -> Optional[str]:
        elem = self.select_one(css)

        if elem is None:
            if strict:
                raise SelectionError(css)

            return None

        return extract(elem, target)

    def scrape(self, css: str, target: Optional[str] = None) -> List[str]:
        output = []

        for elem in self.select(css):
            value = extract(elem, target)

            if value is not None:
                output.append(value)

        return output

    def get_text(self) -> str:
        return super().get_text().strip()

    def get_display_text(self) -> str:
        return get_display_text(self)

    def get_html(self) -> str:
        return self.decode_contents().strip()

    def get_inner_html(self) -> str:
        return self.get_html()

    def get_outer_html(self) -> str:
        return str(self).strip()

    def __getitem__(self, name: str) -> str:
        return cast(str, super().__getitem__(name))

    @overload
    def get(self, name: str, default: str = ...) -> str: ...

    @overload
    def get(self, name: str, default: None = ...) -> Optional[str]: ...

    def get(self, name: str, default: Optional[str] = None) -> Optional[str]:
        return cast(Optional[str], super().get(name, default))

    def get_list(self, name: str) -> List[str]:
        value = super().get(name)

        if not value:
            return []

        return WHITESPACE_RE.split(cast(str, value).strip())


WONDERFUL_ELEMENT_CLASSES = {Tag: MinetTag}


class WonderfulSoup(BeautifulSoup, MinetTag):
    """
    A subclass of BeautifulSoup with some handy tricks such as selection
    wrangling, scraping utilities and better typings.
    """

    def __init__(
        self,
        markup: str,
        features: str = "lxml",
        parse_only: Optional[SoupStrainer] = None,
    ) -> None:
        super().__init__(
            markup,
            features,
            element_classes=WONDERFUL_ELEMENT_CLASSES,
            parse_only=parse_only,
            multi_valued_attributes=None,
        )  # type: ignore


@contextmanager
def suppress_xml_parsed_as_html_warnings(bypass=False):
    try:
        if XMLParsedAsHTMLWarning is None or bypass:
            yield
        else:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=XMLParsedAsHTMLWarning)
                yield
    finally:
        pass



from typing import Union, Optional, List, Dict, Any

from bs4 import BeautifulSoup, SoupStrainer

from minet.scrape.soup import WonderfulSoup
from minet.scrape.analysis import ScraperAnalysisOutputType

AnyScrapableTarget = Union[str, WonderfulSoup, BeautifulSoup]


class ScraperBase(object):
    fieldnames: Optional[List[str]]
    plural: bool
    tabular: bool
    output_type: Optional[ScraperAnalysisOutputType]
    strainer: Optional[SoupStrainer]

    @property
    def singular(self) -> bool:
        return not self.plural

    def __repr__(self):
        return "<{name} plural={plural} output_type={output_type} strain={strain} fieldnames={fieldnames!r}>".format(
            name=self.__class__.__name__,
            plural=self.plural,
            strain=getattr(self.strainer, "css", None) if self.strainer else None,
            output_type=self.output_type,
            fieldnames=self.fieldnames,
        )

    def __call__(self, html: AnyScrapableTarget, context: Optional[Dict] = None) -> Any:
        raise NotImplementedError

    def items(self, html: AnyScrapableTarget, context: Optional[Dict] = None):
        result = self.__call__(html, context=context)

        if result is None:
            return

        if not self.plural:
            yield result
            return

        yield from result



# =============================================================================
# Minet Custom Exceptions
# =============================================================================
#
# Collection of handy custom exceptions.
#
from typing import Optional


# NOTE: duplicated from `minet.utils` to avoid dependency cycles
def message_flatmap(*messages, sep=" ", end="\n"):
    return sep.join(
        end.join(m for m in message) if not isinstance(message, str) else message
        for message in messages
    )


# Base minet error
class MinetError(Exception):
    def __init__(self, message=None):
        if message:
            message = message_flatmap(message)

        self.message = message or ""

        if not message:
            super().__init__()
        else:
            super().__init__(message)

    def __repr__(self):
        representation = "<" + self.__class__.__name__

        for k in dir(self):
            if k.startswith("_") or k == "args" or "traceback" in k:
                continue

            v = getattr(self, k)

            if v is None:
                continue

            if isinstance(v, BaseException):
                representation += " {k}=<{name}>".format(k=k, name=v.__class__.__name__)
                continue

            if isinstance(v, str) and len(v) > 30:
                v = v[:29] + "…"

            representation += " {k}={v!r}".format(k=k, v=v)

        representation += ">"

        return representation


# General errors
class UnknownEncodingError(MinetError):
    pass


class CouldNotInferEncodingError(MinetError):
    pass


# Cookie errors
class UnknownBrowserError(MinetError):
    def __init__(self, browser: str):
        super().__init__("Unknown browser: %s" % browser)
        self.browser = browser


class CookieGrabbingError(MinetError):
    def __init__(self, browser: str, reason: Exception):
        super().__init__("Could not grab cookies from: %s" % browser)
        self.browser = browser
        self.reason = reason


# Miscellaneous HTTP errors
class InvalidURLError(MinetError):
    def __init__(self, url: str):
        self.url = url
        super().__init__(url)


class InvalidStatusError(MinetError):
    def __init__(self, status: int):
        super().__init__(str(status))
        self.status = status


class HTTPCallbackError(MinetError):
    def __init__(self, reason: Exception):
        super().__init__("HTTPCallbackError: " + str(reason))
        self.reason = reason

    def unwrap(self, allow) -> Exception:
        if not isinstance(self.reason, allow):
            raise self.reason

        return self.reason


class CancelledRequestError(MinetError):
    pass


class FinalTimeoutError(MinetError):
    pass


# Redirection errors
class RedirectError(MinetError):
    pass


class MaxRedirectsError(RedirectError):
    pass


class InfiniteRedirectsError(RedirectError):
    pass


class SelfRedirectError(InfiniteRedirectsError):
    pass


class InvalidRedirectError(RedirectError):
    pass


class BadlyEncodedLocationHeaderError(RedirectError):
    pass


# Definition errors
class DefinitionInvalidFormatError(MinetError):
    pass


# Crawling errors
class CrawlError(MinetError):
    pass


class UnknownSpiderError(CrawlError):
    def __init__(self, spider: str):
        super().__init__(spider)
        self.spider = spider


# Extraction errors
class TrafilaturaError(MinetError):
    def __init__(self, message=None, reason=None):
        super().__init__(message)
        self.reason = reason


# Filesystem errors
class FilenameFormattingError(MinetError):
    def __init__(self, message=None, reason=None, template=None):
        super().__init__(message)
        self.reason = reason
        self.template = template


class GenericModuleNotFoundError(MinetError):
    def __init__(self, path: str, target: str):
        self.path = path
        self.target = target
        super().__init__("module %s not found" % self.path)


class TargetInGenericModuleNotFoundError(MinetError):
    def __init__(self, path: str, target: str, name: str):
        self.path = path
        self.target = target
        self.name = name
        super().__init__("target %s in module %s not found" % (self.name, self.path))


# User Agents error
class UserAgentsUpdateError(MinetError):
    def __init__(self, message=None, reason: Optional[BaseException] = None):
        super().__init__(message)
        self.reason = reason


# Pycurl errors
class PycurlNotInstalledError(MinetError):
    pass


class PycurlError(MinetError):
    def __init__(self, reason):
        self.reason = reason
        self.code = reason.args[0]
        super().__init__("%s (%i)" % (reason.args[1], reason.args[0]))


class PycurlTimeoutError(PycurlError):
    pass


class PycurlSSLError(PycurlError):
    pass


class PycurlProtocolError(PycurlError):
    pass


# NOTE: we cannot distinguish connexion error and unknown host errors
# This is the reason why `PycurlHostResolutionError` inherits from
# `PycurlProtocolError` so we can retry it.
class PycurlHostResolutionError(PycurlProtocolError):
    pass


class PycurlConnectionRefusedError(PycurlProtocolError):
    pass


class PycurlReceiveError(PycurlProtocolError):
    pass


class PycurlSendError(PycurlProtocolError):
    pass


# sqlar
class SQLArchiveError(MinetError):
    pass


class SQLArchiveInvalidError(SQLArchiveError):
    pass


# Browser emulation
class BrowserError(MinetError):
    pass


class BrowserYetUnimplementedError(BrowserError):
    pass


class BrowserNameNotResolvedError(BrowserError):
    pass


class BrowserConnectionAbortedError(BrowserError):
    pass


class BrowserConnectionRefusedError(BrowserError):
    pass


class BrowserConnectionClosedError(BrowserError):
    pass


class BrowserConnectionTimeoutError(BrowserError):
    pass


class BrowserTimeoutError(BrowserError):
    pass


class BrowserSSLError(BrowserError):
    pass


class BrowserHTTPResponseCodeFailureError(BrowserError):
    pass


class BrowserContextAlreadyClosedError(BrowserError):
    pass


class BrowserSocketError(BrowserError):
    pass


class BrowserUnableToRetrieveContentError(BrowserError):
    pass



# =============================================================================
# Minet Url Parse CLI Action
# =============================================================================
#
# Logic of the `url-parse` action.
#
from ural import (
    is_url,
    is_shortened_url,
    normalize_url,
    canonicalize_url,
    fingerprint_url,
    get_hostname,
    get_domain_name,
    get_normalized_hostname,
    get_fingerprinted_hostname,
    infer_redirection,
    is_typo_url,
    is_homepage,
    should_resolve,
    could_be_html,
    could_be_rss,
)
from ural.facebook import (
    parse_facebook_url,
    FacebookPost,
    FacebookUser,
    FacebookGroup,
    FacebookHandle,
    FacebookPhoto,
    FacebookVideo,
)
from ural.youtube import (
    parse_youtube_url,
    normalize_youtube_url,
    YoutubeVideo,
    YoutubeUser,
    YoutubeChannel,
    YoutubeShort,
)
from ural.twitter import parse_twitter_url, TwitterTweet, TwitterUser

from minet.cli.utils import with_enricher_and_loading_bar

REPORT_HEADERS = [
    "canonicalized_url",
    "normalized_url",
    "fingerprinted_url",
    "inferred_redirection",
    "domain_name",
    "hostname",
    "normalized_hostname",
    "fingerprinted_hostname",
    "probably_shortened",
    "probably_typo",
    "probably_homepage",
    "should_resolve",
    "could_be_html",
    "could_be_rss",
]

FACEBOOK_REPORT_HEADERS = [
    "facebook_type",
    "facebook_id",
    "facebook_full_id",
    "facebook_handle",
    "facebook_normalized_url",
]

YOUTUBE_REPORT_HEADERS = [
    "youtube_type",
    "youtube_id",
    "youtube_name",
    "youtube_normalize_url",
]

TWITTER_REPORT_HEADERS = ["twitter_type", "twitter_user_screen_name", "tweet_id"]


def extract_standard_addendum(cli_args, url):
    return [
        canonicalize_url(url, quoted=cli_args.quoted),
        normalize_url(
            url,
            infer_redirection=cli_args.infer_redirection,
            fix_common_mistakes=cli_args.fix_common_mistakes,
            normalize_amp=cli_args.normalize_amp,
            quoted=cli_args.quoted,
            sort_query=cli_args.sort_query,
            strip_authentication=cli_args.strip_authentication,
            strip_fragment=cli_args.strip_fragment,
            strip_index=cli_args.strip_index,
            strip_irrelevant_subdomains=cli_args.strip_irrelevant_subdomains,
            strip_protocol=cli_args.strip_protocol,
            strip_trailing_slash=cli_args.strip_trailing_slash,
            platform_aware=cli_args.platform_aware,
        ),
        fingerprint_url(
            url,
            strip_suffix=cli_args.strip_suffix,
            platform_aware=cli_args.platform_aware,
        ),
        infer_redirection(url),
        get_domain_name(url),
        get_hostname(url),
        get_normalized_hostname(
            url,
            infer_redirection=cli_args.infer_redirection,
            normalize_amp=cli_args.normalize_amp,
        ),
        get_fingerprinted_hostname(url, strip_suffix=cli_args.strip_suffix),
        "yes" if is_shortened_url(url) else "no",
        "yes" if is_typo_url(url) else "no",
        "yes" if is_homepage(url) else "no",
        "yes" if should_resolve(url) else "no",
        "yes" if could_be_html(url) else "no",
        "yes" if could_be_rss(url) else "no",
    ]


YOUTUBE_TYPES = {
    YoutubeVideo: "video",
    YoutubeUser: "user",
    YoutubeChannel: "channel",
    YoutubeShort: "short",
}


def extract_youtube_addendum(url):
    parsed = parse_youtube_url(url)

    if parsed is None:
        return None

    return [
        YOUTUBE_TYPES.get(type(parsed)),
        parsed.id,
        getattr(parsed, "name", ""),
        normalize_youtube_url(url),
    ]


def extract_facebook_addendum(url):
    parsed = parse_facebook_url(url)

    if parsed is None:
        return None

    if isinstance(parsed, FacebookPost):
        return ["post", parsed.id, parsed.full_id or "", "", parsed.url]

    elif isinstance(parsed, FacebookHandle):
        return ["handle", "", "", parsed.handle, parsed.url]

    elif isinstance(parsed, FacebookUser):
        return ["user", parsed.id or "", "", parsed.handle or "", parsed.url]

    elif isinstance(parsed, FacebookGroup):
        return ["group", parsed.id or "", "", parsed.handle or "", parsed.url]

    elif isinstance(parsed, FacebookPhoto):
        return ["photo", parsed.id, "", "", parsed.url]

    elif isinstance(parsed, FacebookVideo):
        return ["video", parsed.id, "", "", parsed.url]

    else:
        raise TypeError("unknown facebook parse result type!")


def extract_twitter_addendum(url):
    parsed = parse_twitter_url(url)

    if parsed is None:
        return None

    if isinstance(parsed, TwitterUser):
        return ["user", parsed.screen_name, ""]

    elif isinstance(parsed, TwitterTweet):
        return ["tweet", parsed.user_screen_name, parsed.id]

    else:
        raise TypeError("unknown twitter parse result type!")


def get_headers(cli_args):
    if cli_args.facebook:
        return FACEBOOK_REPORT_HEADERS
    elif cli_args.youtube:
        return YOUTUBE_REPORT_HEADERS
    elif cli_args.twitter:
        return TWITTER_REPORT_HEADERS

    return REPORT_HEADERS


@with_enricher_and_loading_bar(headers=get_headers, title="Parsing", unit="urls")
def action(cli_args, enricher, loading_bar):
    for row, url in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step():
            url = url.strip()

            if not is_url(url, allow_spaces_in_path=True, require_protocol=False):
                enricher.writerow(row)
                continue

            if cli_args.facebook:
                addendum = extract_facebook_addendum(url)
            elif cli_args.youtube:
                addendum = extract_youtube_addendum(url)
            elif cli_args.twitter:
                addendum = extract_twitter_addendum(url)
            else:
                addendum = extract_standard_addendum(cli_args, url)

            if addendum is None:
                enricher.writerow(row)
                continue

            enricher.writerow(row, addendum)



# =============================================================================
# Minet Twitter List Members CLI Action
# =============================================================================
#
# Logic of the `tw list-members` action.
#
import re
from ebbe import getpath
from twitter import TwitterHTTPError
from twitwi import normalize_user, format_user_as_csv_row
from twitwi.constants import USER_PARAMS, USER_FIELDS
from ural.twitter import parse_twitter_url, TwitterList, TwitterTweet, TwitterUser

from minet.cli.twitter.utils import with_twitter_client
from minet.cli.utils import with_enricher_and_loading_bar

ID_RE = re.compile(r"^[0-9]+$")

ITEMS_PER_PAGE = 100


@with_enricher_and_loading_bar(
    headers=USER_FIELDS,
    title="Retrieving list members",
    unit="lists",
    nested=True,
    sub_unit="members",
)
@with_twitter_client(api_version="2")
def action(cli_args, client, enricher, loading_bar):
    for row, twitter_list in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step(twitter_list):
            twitter_id_list = twitter_list

            list_parsed = parse_twitter_url(twitter_list)
            if isinstance(list_parsed, TwitterList):
                twitter_id_list = list_parsed.id
            elif isinstance(list_parsed, (TwitterUser, TwitterTweet)) or (
                not list_parsed and not ID_RE.match(twitter_list)
            ):
                loading_bar.print("%s is not a list id or url." % twitter_list)
                continue

            kwargs = {"max_results": ITEMS_PER_PAGE, "params": USER_PARAMS}

            while True:
                try:
                    result = client.call(
                        ["lists", twitter_id_list, "members"], **kwargs
                    )
                except TwitterHTTPError as e:
                    if e.e.code == 404:
                        enricher.writerow(row)
                    else:
                        raise e

                    break

                if result.get("errors"):
                    loading_bar.print(getpath(result, ["errors", 0, "detail"]))
                    break

                if "data" not in result or result["meta"]["result_count"] == 0:
                    break

                for user in result["data"]:
                    user = normalize_user(user, v2=True)
                    user_row = format_user_as_csv_row(user)
                    enricher.writerow(row, user_row)
                    loading_bar.nested_advance()

                if "next_token" in result["meta"]:
                    kwargs["pagination_token"] = result["meta"]["next_token"]
                else:
                    break



# =============================================================================
# Minet YouTube Constants
# =============================================================================
#
# General constants used throughout the YouTube functions.
#

YOUTUBE_API_BASE_URL = "https://www.googleapis.com/youtube/v3"
YOUTUBE_API_MAX_VIDEOS_PER_CALL = 50
YOUTUBE_API_MAX_CHANNELS_PER_CALL = 50
YOUTUBE_API_MAX_COMMENTS_PER_CALL = 100

YOUTUBE_API_SEARCH_ORDERS = {
    "relevance",
    "date",
    "rating",
    "viewCount",
    "title",
    "videoCount",
}

YOUTUBE_API_DEFAULT_SEARCH_ORDER = "relevance"

# From: https://stackoverflow.com/questions/17698040/youtube-api-v3-where-can-i-find-a-list-of-each-videocategoryid
YOUTUBE_API_CATEGORIES = {
    "1": "Film & Animation ",
    "2": "Autos & Vehicles",
    "10": "Music",
    "15": "Pets & Animals",
    "17": "Sports",
    "18": "Short Movies",
    "19": "Travel & Events",
    "20": "Gaming",
    "21": "Videoblogging",
    "22": "People & Blogs",
    "23": "Comedy",
    "24": "Entertainment",
    "25": "News & Politics",
    "26": "Howto & Style",
    "27": "Education",
    "28": "Science & Technology",
    "29": "Nonprofits & Activism",
    "30": "Movies",
    "31": "Anime/Animation",
    "32": "Action/Adventure",
    "33": "Classics",
    "34": "Comedy",
    "35": "Documentary",
    "36": "Drama",
    "37": "Family",
    "38": "Foreign",
    "39": "Horror",
    "40": "Sci-Fi/Fantasy",
    "41": "Thriller",
    "42": "Shorts",
    "43": "Shows",
    "44": "Trailers",
}



# =============================================================================
# Minet CrowdTangle Leaderboard
# =============================================================================
#
# Function related to leaderboards
#
from minet.crowdtangle.utils import make_paginated_iterator
from minet.crowdtangle.types import CrowdTangleLeaderboard

URL_TEMPLATE = "https://api.crowdtangle.com/leaderboard?count=100&token=%s"


def url_forge(token=None, list_id=None, start_date=None, **kwargs):
    base_url = URL_TEMPLATE % token

    if list_id:
        base_url += "&listId=%s" % list_id

    if start_date is not None:
        base_url += "&startDate=%s" % start_date

    return base_url


crowdtangle_leaderboard = make_paginated_iterator(
    url_forge,
    item_key="accountStatistics",
    item_id_getter=lambda x: x["account"]["id"],
    formatter=CrowdTangleLeaderboard.from_payload,
)



import casanova

from minet.user_agents.data import USER_AGENTS


def action(cli_args):
    writer = casanova.writer(cli_args.output, fieldnames=["user_agent", "percent"])

    for pct, ua in USER_AGENTS:
        writer.writerow([ua, pct])



from typing import List, Any

import casanova
from collections import Counter

from minet.crawl import CrawlResult
from minet.hyphe import HypheSpider, HypheSpiderAddendum

from minet.cli.console import console
from minet.cli.crawl.crawl import crawl_action


def action(cli_args):
    spider = HypheSpider(ignore_internal_links=cli_args.ignore_internal_links)

    reader = casanova.reader(cli_args.corpus)
    headers = reader.headers

    assert headers is not None

    id_pos = headers[cli_args.id_column]
    status_pos = headers[cli_args.status_column]
    prefixes_pos = headers[cli_args.prefixes_column]
    start_pages_pos = headers.get(cli_args.start_pages_column)

    status_counts = Counter()
    webentity_count = 0

    for row in reader:
        webentity = row[id_pos]
        status = row[status_pos] or "IN"
        prefixes = row[prefixes_pos]

        if not prefixes:
            continue

        webentity_count += 1
        status_counts[status] += 1

        for prefix in prefixes.split(cli_args.prefix_separator):
            spider.set(prefix, webentity, status=status)

            if status == "IN":
                spider.add_start_page(prefix)

        if status != "IN" or start_pages_pos is None:
            continue

        start_pages = row[start_pages_pos]

        if not start_pages:
            continue

        for start_page in start_pages.split(cli_args.start_page_separator):
            spider.add_start_page(start_page)

    # Report
    console.print(
        "Indexed [success]%i[/success] webentities through [success]%i[/success] prefixes."
        % (webentity_count, len(spider.trie))
    )
    console.print("  [success]IN[/success]: %i" % status_counts["IN"])
    console.print("  [error]OUT[/error]: %i" % status_counts["OUT"])
    console.print("  [dim]UNDECIDED[/dim]: %i" % status_counts["UNDECIDED"])
    console.print("  [info]DISCOVERED[/info]: %i" % status_counts["DISCOVERED"])
    console.print(
        "Registered [success]%i[/success] start pages" % len(spider.start_pages)
    )

    def format_job_row_addendum(result: CrawlResult[Any, HypheSpiderAddendum]) -> List:
        info = result.data

        if not info:
            url = result.job.url
            webentity = spider.trie.match(url)

            assert webentity is not None

            return [webentity.id]

        return [info.webentity_id]

    return crawl_action(
        cli_args,
        target=spider,
        additional_job_fieldnames=["webentity"],
        format_job_row_addendum=format_job_row_addendum,
    )



# =============================================================================
# Minet Twitter Tweets CLI Action
# =============================================================================
#
# Logic of the `tw tweets` action.
#
from casanova import Enricher
from twitwi.constants import TWEET_FIELDS
from twitwi import format_tweet_as_csv_row
from ural.twitter import is_twitter_url, parse_twitter_url, TwitterTweet

from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.loading_bar import LoadingBar
from minet.twitter import TwitterGuestAPIScraper
from minet.twitter.exceptions import TwitterPublicAPINotFound


@with_enricher_and_loading_bar(
    headers=TWEET_FIELDS,
    title="Scraping",
    unit="tweets",
)
def action(cli_args, enricher: Enricher, loading_bar: LoadingBar):
    scraper = TwitterGuestAPIScraper()

    for row, tweet_id_or_url in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step():
            tweet_id = tweet_id_or_url

            if is_twitter_url(tweet_id_or_url):
                parsed = parse_twitter_url(tweet_id_or_url)

                if isinstance(parsed, TwitterTweet):
                    tweet_id = parsed.id

            try:
                tweet = scraper.tweet(tweet_id)
                tweet = format_tweet_as_csv_row(tweet)
                enricher.writerow(row, tweet)
            except TwitterPublicAPINotFound:
                loading_bar.inc_stat("not-found", style="warning")
                enricher.writerow(row)
                continue



from typing import Optional, List, Any

from bs4 import SoupStrainer, BeautifulSoup

from minet.scrape.analysis import ScraperAnalysisOutputType
from minet.scrape.utils import ensure_soup
from minet.scrape.types import AnyScrapableTarget
from minet.scrape.classes.base import ScraperBase


class NamedScraper(ScraperBase):
    name: str
    fieldnames: List[str]
    plural: bool
    tabular = True
    output_type: ScraperAnalysisOutputType
    strainer: Optional[SoupStrainer]

    def scrape(self, soup: BeautifulSoup, context=None) -> Any:
        raise NotImplementedError

    def __call__(self, html: AnyScrapableTarget, context=None) -> Any:
        soup = ensure_soup(html, strainer=self.strainer)
        return self.scrape(soup, context=context)



# =============================================================================
# Minet BuzzSumo Domain Summary CLI Action
# =============================================================================
#
# Logic of the `bz domain-summary` action.
#
from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.buzzsumo.utils import with_buzzsumo_fatal_errors
from minet.buzzsumo import BuzzSumoAPIClient

SUMMARY_HEADERS = ["total_results", "total_pages"]


@with_buzzsumo_fatal_errors
@with_enricher_and_loading_bar(
    title="Retrieving domain summary", headers=SUMMARY_HEADERS, unit="domains"
)
def action(cli_args, enricher, loading_bar):
    client = BuzzSumoAPIClient(cli_args.token)

    for row, domain_name in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step():
            data = client.domain_summary(
                domain_name, cli_args.begin_date, cli_args.end_date
            )

            enricher.writerow(row, [data["total_results"], data["total_pages"]])



# =============================================================================
# Minet CrowdTangle Constants
# =============================================================================
#
# General constants used throughout the CrowdTangle functions.
#
from urllib3 import Timeout

CROWDTANGLE_DEFAULT_RATE_LIMIT = 6  # Number of hits per minute
CROWDTANGLE_LINKS_DEFAULT_RATE_LIMIT = 2
CROWDTANGLE_DEFAULT_START_DATE = "2010"

CROWDTANGLE_DEFAULT_TIMEOUT = Timeout(connect=10, read=60 * 5)

CROWDTANGLE_POST_TYPES = [
    "episode",
    "extra_clip",
    "link",
    "live_video",
    "live_video_complete",
    "live_video_scheduled",
    "native_video",
    "photo",
    "status",
    "trailer",
    "tweet",
    "video",
    "vine",
    "youtube",
]

CROWDTANGLE_SORT_TYPES = [
    "date",
    "interaction_rate",
    "overperforming",
    "total_interactions",
    "underperforming",
]

CROWDTANGLE_SUMMARY_SORT_TYPES = {"date", "subscriber_count", "total_interactions"}

CROWDTANGLE_SUMMARY_DEFAULT_SORT_TYPE = "date"

CROWDTANGLE_SEARCH_FIELDS = {
    "text_fields_and_image_text",
    "include_query_strings",
    "text_fields_only",
    "account_name_only",
    "image_text_only",
}

CROWDTANGLE_DEFAULT_SEARCH_FIELD = "text_fields_and_image_text"

CROWDTANGLE_STATISTICS = [
    "like",
    "share",
    "favorite",
    "comment",
    "love",
    "wow",
    "haha",
    "sad",
    "angry",
    "thankful",
    "care",
]

CROWDTANGLE_FULL_STATISTICS = STATISTICS = [
    ("loveCount", "love_count"),
    ("wowCount", "wow_count"),
    ("thankfulCount", "thankful_count"),
    ("interactionRate", "interaction_rate"),
    ("likeCount", "like_count"),
    ("hahaCount", "haha_count"),
    ("commentCount", "comment_count"),
    ("shareCount", "share_count"),
    ("sadCount", "sad_count"),
    ("angryCount", "angry_count"),
    ("postCount", "post_count"),
    ("totalInteractionCount", "total_interaction_count"),
    ("totalVideoTimeMS", "total_video_time_ms"),
    ("threePlusMinuteVideoCount", "three_plus_minute_video_count"),
]

CROWDTANGLE_PLATFORMS = {"facebook", "instagram", "reddit"}

CROWDTANGLE_REACTION_TYPES = [
    "angry",
    "comment",
    "haha",
    "like",
    "love",
    "sad",
    "share",
    "thankful",
    "wow",
]



from typing import Any

import warnings
from datetime import datetime
from html import unescape
from bs4 import SoupStrainer, BeautifulSoup, MarkupResemblesLocatorWarning

from .types import NamedScraper


def extract_content(content):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=MarkupResemblesLocatorWarning)
        return BeautifulSoup(unescape(content), "html.parser").get_text().strip()


def extract_date(doc_id):
    return datetime.strptime(doc_id.split("·")[1], "%Y%m%d").date().isoformat()


def extract_media(media):
    return media.split(",", 1)[0].split("\n", 1)[0].split(" " * 16, 1)[0].strip()


def select_and_strip(elem, selector):
    selected_elem = elem.select_one(selector)

    if selected_elem is None:
        return ""

    return selected_elem.get_text().strip()


class EuropresseScraper(NamedScraper):
    name = "europresse"
    fieldnames = ["id", "title", "content", "url", "date", "media", "media_id"]
    plural = True
    output_type = "collection"
    strainer = SoupStrainer(name="article")

    def scrape(self, soup: BeautifulSoup, context=None) -> Any:
        articles = []
        selectors = {
            "title": ".titreArticle",
            "id": ".publiC-lblNodoc",
            "media": ".DocPublicationName",
        }

        for elem in soup.select("article"):
            row = {}

            content = elem.select_one(".docOcurrContainer")
            if content is None:
                content = ""
            else:
                urls = content.select("a")
                for u in urls:
                    if "Cet article est paru dans" in u.get_text():
                        row["url"] = u.get("href")
                        break
                content = content.get_text()

            row["content"] = extract_content(content)

            for field, selector in selectors.items():
                row[field] = select_and_strip(elem, selector)

            row["date"] = extract_date(row["id"])
            row["media"] = extract_media(row["media"])
            row["media_id"] = row["id"].split("·")[2]

            articles.append(row)

        return articles



# =============================================================================
# Minet Tiktok Search-videos CLI Action
# =============================================================================
#
# Logic of the `tiktok search-videos` action.
#
from itertools import islice

from minet.cli.utils import with_enricher_and_loading_bar
from minet.tiktok import TiktokAPIScraper
from minet.tiktok.types import TiktokVideo


@with_enricher_and_loading_bar(
    headers=TiktokVideo,
    title="Searching videos",
    unit="queries",
    nested=True,
    sub_unit="videos",
)
def action(cli_args, enricher, loading_bar):
    client = TiktokAPIScraper(cookie=cli_args.cookie)

    for row, query in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step(query):
            generator = client.search_videos(query)

            if cli_args.limit:
                generator = islice(generator, cli_args.limit)

            for video in generator:
                enricher.writerow(row, video.as_csv_row())
                loading_bar.nested_advance()
